"""
Work with vector data and attributes with ease.
"""
from .core_vector import *
from .clip import *
from .convert_parts import *
from .dissolve import *
from .grid import *
from .intersect import *
from .merge import *
from .rasterize import *
from .reproject import *
from .buffer import *
from .split import *
from .metadata import *
from .shapes import *
