"""
Create vector grids from references or hardcode
"""

# TODO: Implement grid functions
# TODO: create_grid, create_grid_points, create_grid_linestrings

def _vector_create_grid():
    """ Not implemented."""
    raise RuntimeError("Not implemented")
